#include <math.h>
//#include "matrix.h"
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include "ndim_interpolator.h"
#include "rungekutta.h"

double Qflow=1;
int a=1;
ndim_interpolator* compc;

// Holds the thermodynamic state of a gas
class gasstate{
public:
  double p,v,t;

  friend ostream& operator<<(ostream&,const gasstate*);
  friend ofstream& operator<<(ofstream&,const gasstate*);
  friend ifstream& operator>>(ofstream&,gasstate*);
};

// Holds the states for the different stages of the engine
// (intake, after compressor, after
class jetstate {
public:
  gasstate s0,s1,s2,s3,s4;

  friend ostream& operator<<(ostream&,const jetstate*);  
  friend ofstream& operator<<(ofstream&,const jetstate*);  
  friend ifstream& operator>>(ofstream&,jetstate*);  
};

// Auxiliary (non-thermodynamic) state variables.
class jetenvirons {
public:
  double p0,t0;
  double a0,a1,a2,a3,a4;       // cross-sectional areas
  double omega,refrpm;         // shaft angular velocity reference
  double mu,g,cp;              // gas parameters 
  double Q;                    // combustor heat input
  static double R;
  double Ishaft;               // compressor/turbine moment of inertia
  double tau;                  // shaft torque
  double combPloss;
  double eta_compr;            // efficiencies
  double eta_turbine;
  double alfa,beta;
  double setthrottle;
};

double jetenvirons::R=8.31;

// Return "total" pressure
double ptotal(gasstate s,jetenvirons env)
{
  return s.p*(1+s.v*s.v*env.mu/(2*s.t*env.R));
}

// Return "total" temperature
double ttotal(gasstate s,jetenvirons env)
{
  return s.t+s.v*s.v/(2*env.cp);
}

// mass rate
double mdot(gasstate j,double a, jetenvirons env)
{
  return a*j.v*j.p*env.mu/(j.t*env.R);
}

// The Runge-Kutta is only used for integrating the shaft rpm (the
// thermodynamic equations are solved iteratively for every timestep)
class jetintegrator : public RKintegrator {
protected:
  matrix<double> dydxfunction(double,matrix<double>);
  void solvestate(matrix<double>);
  jetstate iterateState(jetstate,const jetenvirons);
public:
  jetstate j;
  jetenvirons e;

  jetintegrator(double,matrix<double>,double,matrix<double>,double);
};

jetintegrator::jetintegrator(double a,matrix<double> b,double c,matrix<double> d,double f) : RKintegrator(a,b,c,d,f)
{
  e.beta=0.5;
  e.alfa=1.1;
  e.setthrottle=1;
}

// Right-hand side of the equations
// y[0]=throttle
// y[1]=dthrottle/dt
// y[2]=rpm
matrix<double> jetintegrator::dydxfunction(double t, matrix<double> y)
{
  matrix<double> yout(y.getrows(),1,0.);
  
  // First we must solve for the jet thermodynamic state for this y
  solvestate(y);

  yout[0] = y[1];
  yout[1] = -e.alfa*y[1] - e.beta*(y[0]-e.setthrottle);
  yout[2] = e.tau/e.Ishaft;
  
  return yout;
}

// Solves for the jet state specified by throttle/rpm in y by straightforward
// iteration. (It's a reasonably complicated system of eqs)
void jetintegrator::solvestate(matrix<double> y)
{
  e.Q     = y[0]*21.9e6;
  e.omega = y[2];

  j= iterateState(j,e);
  j= iterateState(j,e);
  j= iterateState(j,e);
  double oldv=0;
  int n=0;
  while((fabs(oldv-j.s0.v)>0.001)&&n<300) {
    //    cout << in << '\n';
    //    scanf("%s",slask);
    oldv=j.s0.v;
    j= iterateState(j,e);
    n++;
  }

  if(n>=300)
    cerr << "NOT CONVERGING!!\n";

  // Now that we have the thermodynamic state, we can calculate the
  // net shaft torque
  double Wc=e.cp*(j.s1.t-j.s0.t)+(pow(j.s1.v,2)-pow(j.s0.v,2))/2;
  double Wt=e.cp*(j.s2.t-j.s3.t)+(pow(j.s2.v,2)-pow(j.s3.v,2))/2;
  e.tau=(Wt-Wc)*mdot(j.s2,e.a2,e)/e.omega;
  //e.tau=(Wt*0.9-Wc/0.85)*mdot(j.s2,e.a2,e)/e.omega;
}

/*
jetstate solvestate(jetstate in,jetenvirons e)
{
  in= iterateState(in,e);
  in= iterateState(in,e);
  in= iterateState(in,e);
  double oldv=0;
  int n=0;
  while((fabs(oldv-in.s0.v)>0.001)&&n<300) {
    //    cout << in << '\n';
    //    scanf("%s",slask);
    oldv=in.s0.v;
    in= iterateState(in,e);
    n++;
  }
  if(n>=300)
    cerr << "NOT CONVERGING!!\n";
  return in;
}
*/

jetintegrator* ji;

static void Key(unsigned char key,int x,int y)
{
  switch(key){
  case 27:
    a=0;
    exit(0);
    break;
  case '+':
    ji->e.setthrottle+=0.01;
    break;
  case '-':
    ji->e.setthrottle-=0.01;
    break;
  case '1':
    ji->e.setthrottle=1;
    break;
  case '0':
    ji->e.setthrottle=0;
    break;
    }
glutPostRedisplay();
}

static void display()
{
glClear(GL_COLOR_BUFFER_BIT);
glutSwapBuffers();
}

static void reshape(int a,int b)
{
}

// Returns compressor total pressure ratio from corrected rpm and mass flow
// This is the real deal - we need to put authentic data in here
double compressorchar(gasstate s, jetenvirons env)
{
  double sqtemp = sqrt(ttotal(s,env));
  double caf=mdot(s,env.a0,env)*sqtemp/ptotal(s,env)/0.004849;
  double crpm=env.omega/sqtemp/583.701;

  // Now we do it with the interpolator and an arbitrary scaling
  double pr = 14.0/6*compc->interpolate(caf/1.0*7.2,crpm*0.9);

  return pr;
}

// Returns turbine total pressure ratio from corrected rpm and mass flow
// This is the real deal - we need to put authentic data in here
double turbinechar(gasstate s, jetenvirons env)
{
  double sqtemp = sqrt(ttotal(s,env));
  double caf=mdot(s,env.a2,env)*sqtemp/ptotal(s,env)/0.000712;
  double crpm=env.omega/sqtemp/282.226;

  //double crpmfactor=pow(crpm,0.9);
  //double pr=(3-0.2*(caf/crpmfactor))*crpmfactor;
  double dx=2-2*exp(-crpm/4);
  double dy=1.2-1.2*exp(-crpm/4);
  double pr=3-0.2*pow((caf-dx)/2,4)+dy;
  // Fudge a simple characteristic
  return pr;//3-250*caf;
}

// Performs one iteration of trying to solve the state equations
// in contains the current state variables, env the environmental ones
jetstate jetintegrator::iterateState(jetstate in, const jetenvirons env)
{
  double g = env.g; // gamma - will be optimized out anyway
  const double muR = env.mu/env.R;

  jetstate out=in;
  in.s0.p = env.p0;
  in.s0.t = env.t0;

  // intake
  out.s0.p = in.s0.p;
  out.s0.t = in.s0.t;

  // compressor
  // compute the gamma of the process based on the pressure ratio and the
  // efficiency
  double cpr=compressorchar(in.s0,env);
  double ce=env.eta_compr;
  double cg=g/(1+g*log(ce)/log(cpr));

  out.s1.v = env.a0*in.s1.t*in.s0.p*in.s0.v/(env.a1*in.s0.t*in.s1.p);
  out.s1.t = in.s0.t*pow(in.s1.p/in.s0.p,(cg-1)/cg);
  out.s1.p = ptotal(in.s0,env)/(1+pow(in.s1.v,2)*muR/(2*in.s1.t)) * cpr;

  // combustor
  out.s2.v = env.a1*in.s2.t*in.s1.p*in.s1.v/(env.a2*in.s1.t*in.s2.p);
  out.s2.p = in.s1.p*env.combPloss;
  //out.s2.t = in.s1.t + ( pow(in.s1.v,2)/2 - pow(in.s2.v,2)/2 + env.Q )/env.cp;
  // Q is total heat which is divided by the current mass flow
  out.s2.t = in.s1.t + ( pow(in.s1.v,2)/2 - pow(in.s2.v,2)/2 + 
			 env.Q*in.s1.t/(env.a1*in.s1.v*in.s1.p*muR) )/env.cp;
  if(out.s2.t<in.s1.t) out.s2.t=in.s1.t;
  
  // turbine
  out.s3.v = env.a2*in.s3.t*in.s2.p*in.s2.v/(env.a3*in.s2.t*in.s3.p);
  out.s3.p = ptotal(in.s2,env)/( 1+pow(in.s3.v,2)*muR/(2*in.s3.t) ) /
    turbinechar(in.s2,env);
  out.s3.t = in.s2.t*pow(out.s3.p/in.s2.p,(g-1)/g);
  //out.s3.t = in.s2.t*pow(in.s3.p/in.s2.p,(g-1)/g);
 
  // exhaust
  out.s4.p = in.s0.p;
  out.s4.t = in.s3.t*pow(in.s4.p/in.s3.p,(g-1)/g);
  double temp = 2*(pow(in.s3.v,2)/2 + env.cp*(in.s3.t-in.s4.t) );
  out.s4.v = (temp>0)?sqrt(temp):1;
  
  // close the loop - slowly so it converges
  out.s0.v = in.s0.v+
    (env.a4*in.s0.t*in.s4.p*in.s4.v/(env.a0*in.s4.t*in.s0.p)-in.s0.v)/100;

  return out;
}



ostream& operator<<(ostream& s, const gasstate& g)
{
  s << '(' << g.p << '\t' << g.v << '\t' << g.t << "\t)";
  return s;
}

ofstream& operator<<(ofstream& s, const gasstate& g)
{
  s << g.p << '\t' << g.v << '\t' << g.t << '\n';
  return s;
}

ifstream& operator>>(ifstream& s, gasstate& g)
{
  s >> g.p >> g.v >> g.t;
  return s;
}

ostream& operator<<(ostream& s, const jetstate& g)
{
  s << "0:\t" << g.s0 << '\n';
  s << "1:\t" << g.s1 << '\n';
  s << "2:\t" << g.s2 << '\n';
  s << "3:\t" << g.s3 << '\n';
  s << "4:\t" << g.s4 << '\n';
  return s;
}

ofstream& operator<<(ofstream& s, const jetstate& g)
{
  s << g.s0 << g.s1 << g.s2 << g.s3 << g.s4;
  return s;
}

ifstream& operator>>(ifstream& s, jetstate& g)
{
  s >> g.s0 >> g.s1 >> g.s2 >> g.s3 >> g.s4;
  return s;
}

double printint=1;
double lastx=0;

void calculate()
{
  ji->RKadvance();
  ji->solve(ji->getx()+0.5);
    

  // Time to dump?
    if((ji->getx()-lastx)/printint>1) {
      double thrust=mdot(ji->j.s0,ji->e.a0,ji->e)*(ji->j.s4.v-ji->j.s0.v);
      double massflow=mdot(ji->j.s0,ji->e.a0,ji->e);
      double Wc=ji->e.cp*(ji->j.s1.t-ji->j.s0.t)+
	(pow(ji->j.s1.v,2)-pow(ji->j.s0.v,2))/2;
      double Wt=ji->e.cp*(ji->j.s2.t-ji->j.s3.t)+
	(pow(ji->j.s2.v,2)-pow(ji->j.s3.v,2))/2;

      cerr.precision(1);
      cerr.width(5);
      cerr.setf(ios::scientific,ios::floatfield);
      cerr << ji->getx() << "\t" << ji->e.setthrottle << "\t" 
	   << ji->gety()[0] << '\t' << ji->e.omega 
	   << "\t" << ji->j.s3.t
	   << "\t" << massflow << "\t" << thrust << "\t"
	   << ji->e.tau << "\t" 
	   << ptotal(ji->j.s1,ji->e)/ptotal(ji->j.s0,ji->e) << "\t"
	   << ji->j.s4.v<< "\r";

      cout << ji->getx() << '\t' << ji->e.setthrottle << '\t' 
	   << ji->gety()[0] << '\t' << ji->e.omega 
	   << '\t' << ji->j.s3.t
	   << '\t' << massflow << '\t' << thrust << '\t'
	   << ji->e.tau << "\t" 
	   << ptotal(ji->j.s1,ji->e)/ptotal(ji->j.s0,ji->e) << "\t"
	   << ji->j.s4.v<< "\n";

    }

    //usleep(1000000/2);
    //}
}

main(int argc,char** argv)
{  
  
  // Open a glut window for key i/o
  glutInit( &argc, argv );
  glutInitWindowPosition( 600, 0 );
  glutInitWindowSize( 400, 400 );  
  glutInitDisplayMode( GLUT_RGB);  
  glutCreateWindow(argv[0]);
  glutKeyboardFunc( Key );
  glutIdleFunc(calculate);
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  

  matrix<double> init(3,1,1.,0.,1e4);
  matrix<double> scale(3,1,1.,1.,1e4);

  ji = new jetintegrator(0,init,0.001,scale,0.1);
  ji->setmaxstep(0.5);

  // set up environmental variables
  ji->e.p0=101e3;
  ji->e.t0=288;
  
  ji->e.a0=.24;
  ji->e.a1=0.035;
  ji->e.a2=0.025;
  ji->e.a3=0.07;
  ji->e.a4=.06;
  
  ji->e.refrpm=350;
  ji->e.omega=1e4;
  ji->e.Ishaft=10;
  ji->e.mu=2.88e-2;
  ji->e.g=1.40;
  ji->e.cp=1001.4;
  ji->e.Q=21.9e6;
  ji->e.combPloss=0.95;
  ji->e.eta_compr=0.85;
  ji->e.eta_turbine=0.9;
  
  // Load Compressor characteristics interpolator data
  compc = new ndim_interpolator("compchar.dat");

  // Read it from file
  ifstream inf("jetstats");
  inf >> ji->j;

  cerr << "Time\tThrot\tOmega\tTIT\tMdot\tThrust\tTorque\tComprPr\tExh.vel\n";

  glutMainLoop();
  while(1)
    calculate();


}

